package org.garret.serme;

import java.util.Enumeration;

/**
 * Interface of persistent storage. Programmer should use instance of one of particular 
 * implementations if this unterface. <code>IStorage</code> interface is needed when the application needs
 * to perfrom manipulations with persistent data (add/update/delete objects) without loading/saveing the
 * whole object three to/from memory. If application needs only serialization mechanism, <code>IStorage</code>
 * interface is not needed.
 */
public interface IStorage {
    /**
     * Store object in the database. 
     * If object is not yet persistent (not assigned OID), then it is made persistent.
     * When object is stored all transient objects (objets without assigned OID) referenced from this object
     * are recursively stored in the database. But referenced persistent objects are not recursively stored. 
     * So you should exeplicitly invoke <code>store</code> method for each modified persistent object.
     */
    void store(Object obj);
    
    /**
     * Remove object from the storage. All references to the removed object will be replaced with 
     * null when database will be opened next time. 
     */
    void remove(Object obj);
    /**
     * Open existed or create new database
     * @return <code>true</code> if database was successfully opened, <code>false</code> otherwise
     */
    boolean create();
    /**
     * Open existed database. Method will fail if database doesn't exist.
     * @return <code>true</code> if database was successfully opened, <code>false</code> otherwise
     */
    boolean open();
    /**
     * Remove database. This method can be invoked only for not opened database.
     * @return <code>true</code> if database was successfully deleted, <code>false</code> otherwise
     */
    boolean remove();
    /**
     * Close database
     */
    void close();
    /**
     * Get root object. Database can have only one root object. When database is just created, there
     * are not root object at all and this method returns null
     */
    Object getRoot();
    /**
     * Set new root object. Old root object is not removed from the database.
     * @param root new root object. If it is not yet persistent (has no assigned OID), it is automatically
     * made persistent and stored in the database.
     */
    void setRoot(Object root);
    /**
     * Get enumeration of all live persistent objects in the database. Objects are returned in arbitrary order.
     * @return enumeration of all persistent objects in the database
     */
    Enumeration enumerateObjects();
}