package org.garret.serme;

import java.util.*;
import org.garret.serme.impl.*;

/**
 * Indexed collection providing fast access and sorting of their members. 
 * Members of the collection are sorted using provided user defined comparator.
 * The instance of comparator class is also stored in database and provides 
 * methods for comparing collection members with each other and with key.
 * SortedCollection is implemented using T-Tree, optimizing both speed of main
 * operations (insert/find/remove) and space used by collection in memory. 
 */
public class SortedCollection extends Persistent {
    private IComparator comparator;
    private boolean     unique;
    private TtreePage   root;
    private int         nMembers;
    private int         counter;

    /**
     * Default constructor internally used by SerMe
     */
    public SortedCollection() {} 

    /**
     * Constructor of sorted collection
     * @param comparator class used to compare collection members with each other and with key
     * @param unique if key should be unique in collection
     */
    public SortedCollection(IComparator comparator, boolean unique) { 
        this.comparator = comparator;
        this.unique = unique;
    }

    /**
     * Get comparator used in this collection
     * @return collection comparator
     */
    public IComparator getComparator() { 
        return comparator;
    }

    /*
     * Get member with specified key.
     * @param key specified key. It should match with type of the index and should be inclusive.
     * @return object with this value of the key or <code>null</code> if key not found
     * @exceptionSermeException("Key not unique") exception if there are more than 
     * one objects in the collection with specified value of the key.  
     */
    public Object get(Object key) { 
        if (root != null) { 
            Vector list = new Vector();
            root.find(comparator, key, key, list);
            if (list.size() > 1) { 
                throw new SermeException("Key not unique");
            } else if (list.size() == 0) { 
                return null;
            } else { 
                return list.elementAt(0);
            }
        }
        return null;
    }

            

    /**
     * Get members which key value belongs to the specified range.
     * Either from boundary, either till boundary either both of them can be <code>null</code>.
     * In last case the method returns all objects from the collection.
     * @param from inclusive low boundary. If <code>null</code> then low boundary is not specified.
     * @param till inclusive high boundary. If <code>null</code> then high boundary is not specified.
     * @return array of objects which keys belongs to the specified interval, ordered by key value
     */
    static final Object[] emptySelection = new Object[0];

    public Object[] get(Object from, Object till) { 
        if (root != null) { 
            Vector list = new Vector();
            root.find(comparator, from, till, list);
            if (list.size() != 0) { 
                Object[] mbrs = new Object[list.size()];
                for (int i = 0; i < mbrs.length; i++) { 
                    mbrs[i] = list.elementAt(i);
                }
                return mbrs;
            }
        }
        return emptySelection;
    }


    /**
     * Add new member to collection
     * @param obj new member
     * @return <code>true</code> if object is successfully added in the index, 
     * <code>false</code> if collection was declared as unique and there is already member with such value
     * of the key in the collection. 
     */
    public boolean add(Object obj) { 
        TtreePage newRoot;
        if (root == null) { 
            newRoot = new TtreePage(obj);
        } else { 
            TtreePageReference ref = new TtreePageReference(root);
            if (root.insert(getStorage(), comparator, obj, unique, ref) == TtreePage.NOT_UNIQUE) { 
                return false;
            }
            newRoot = ref.pg;
        }
        root = newRoot;
        nMembers += 1;
        counter += 1;
        store();
        return true;
    }
                
                
    /**
     * Check if collections contains specified member
     * @return <code>true</code> if specified member belongs to the collection
     */
    public boolean contains(Object member) {
        return (root != null)  ? root.contains(comparator, member) : false;
    }        

    /**
     * Remove member from collection
     * @param obj member to be removed
     * @exception StorageError(StorageError.KEY_NOT_FOUND) exception if there is no such key in the collection
     */
    public void remove(Object obj) {
        if (root == null) {
            throw new SermeException("Key not found");
        }
        TtreePageReference ref = new TtreePageReference(root);
        if (root.remove(getStorage(), comparator, obj, ref) == TtreePage.NOT_FOUND) {             
            throw new SermeException("Key not found");
        }
        root = ref.pg;
        nMembers -= 1;        
        counter += 1;
        store();
    }

    /**
     * Get number of objects in the collection
     * @return number of objects in the collection
     */
    public int size() {
        return nMembers;
    }
    
    /**
     * Remove all objects from the collection
     */
    public void clear() {
        if (root != null) { 
            if (getStorage() != null) { 
                root.prune(getStorage());
            }
            root = null;
            nMembers = 0;
            counter += 1;
            store();
        }
    }
 
    /**
     * Get all objects in the index as array ordered by index key.
     * @return array of objects in the index ordered by key value
     */
    public Object[] toArray() {
        if (root == null) { 
            return emptySelection;
        }
        Object[] arr = new Object[nMembers];
        root.toArray(arr, 0);
        return arr;
    }

    /**
     * Get all objects in the index as array ordered by index key.
     * Object are placed in specified array. If number of colelction members
     * is greater than array size, then only first N of them will be stored, where N is 
     * length of array.<BR>
     * If this index fits in the specified array with room to spare
     * (i.e., the array has more elements than this index), the element
     * in the array immediately following the end of the index is set to
     * <tt>null</tt>.  This is useful in determining the length of this
     * index <i>only</i> if the caller knows that this index does
     * not contain any <tt>null</tt> elements.)<p>
     * @return array of objects in the index ordered by key value
     */
    public Object[] toArray(Object[] arr) {
        if (root != null) { 
            root.toArray(arr, 0);
        }
        if (arr.length > nMembers) { 
            arr[nMembers] = null;
        }
        return arr;
    }

    /**
     * Get enumeration for all collection members.
     * Enumertion will be perfromed in order specified by comparator.
     * It is not possible to modify collection duting enumeration - if such 
     * concurrent modification will be detected, then <code>SermeException("Concurrent modification")</code>
     * exception will be thrown.
     * @return enumeration for collection elements
     */
    public Enumeration elements() { 
        return new TtreeEnumeration(this, root);
    }

    /**
     * Pack object
     * @param out output stream
     */
    public void writeObject(IOutputStream out) { 
        out.writeObject(comparator);
        out.writeBoolean(unique);
        out.writeObject(root);
        out.writeInt(nMembers);
    }

    /**
     * Unpack object
     * @param in input stream
     */
    public void readObject(IInputStream in) { 
        comparator = (IComparator)in.readObject();
        unique = in.readBoolean();
        root = (TtreePage)in.readObject();
        nMembers = in.readInt();
    }

    /**
     * This method is used internally by SerMe to detect concurrent modifications
     */
    public int getCounter() { 
        return counter;
    }
}

