package org.garret.serme.file;

import java.util.*;
import java.io.*;
import javax.microedition.io.*;
import javax.microedition.io.file.*;

import org.garret.serme.*;
import org.garret.serme.impl.*;

/**
 * Storage implementation on top of javax.microtedition.io.file package.
 */
public class FileStorage implements IStorage { 
    static final int OBJ_HDR_SIZE = 0;
    static final int OBJ_HDR_OID = 4;
    static final int OBJ_HDR_CID = 8;
    static final int SIZE_OF_OBJ_HDR = 12;
    static final int ALLOC_QUANTUM = 16; // should be power of two and not less than SIZE_OF_OBJ_HDR
    static final int SIZE_OF_DB_HDR = ALLOC_QUANTUM;
    static final int DB_HDR_ROOT = 0;

    /**
     * Storage constructor
     * @param path path to the database file
     */
    public FileStorage(String path) { 
        this.path = path;
    }

    protected void checkIfOpened(boolean shouldBeOpened) { 
        if (opened != shouldBeOpened) { 
            throw new SermeException("Incorrect storage state");
        }
    }

    public void store(Object obj) { 
        checkIfOpened(true);
        output.storeObject(obj);
    }

    public void remove(Object obj) { 
        checkIfOpened(true);
        ObjectHeader hdr = objToHdr.remove(obj);
        if (hdr == null) { 
            throw new SermeException("Object not persistent");
        }
        try { 
            free(hdr.offs, hdr.size);
        } catch (IOException x) { 
            throw new SermeException("File access error: " + x);
        }            
    }

    public boolean open() { 
        return open0(false);
    }

    public boolean create() { 
        return open0(true);
    }

    public boolean remove() { 
        checkIfOpened(true);
        try { 
            con.delete();
        } catch (IOException x) { 
            return false;
        }
        return true;
    }

    static int align(int size) { 
        return (size + ALLOC_QUANTUM - 1) & ~(ALLOC_QUANTUM-1);
    }

    void free(int pos, int size) throws IOException { 
        byte[] hdrBuf = new byte[SIZE_OF_OBJ_HDR];
        Converter.packInt(hdrBuf, OBJ_HDR_SIZE, size);          
        OutputStream out = con.openOutputStream(pos);
        out.write(hdrBuf);                
        out.close();
        addHole(pos, size);
    }

    int allocate(int size) throws IOException { 
        MemoryHole hole = memory; 
        while (true) { 
            hole = hole.next;
            int rest = hole.size - size;
            if (rest == 0) { 
                hole.unlink();
                return hole.offs;
            } else {   
                int pos = hole.offs;
                hole.offs += size;
                hole.size -= size;
                if (hole != memory) { 
                    free(hole.offs, hole.size); 
                }
                return pos;
            }
        }
    }
                
    protected boolean open0(boolean create) {         
        checkIfOpened(false);
        try { 
            String url = path;
            if (!url.startsWith("file:")) { 
                if (url.startsWith("/")) { 
                    url = "file:///" + path;
                } else { 
                    Enumeration e = FileSystemRegistry.listRoots();
                    // choose arbitrary root directory
                    url = "file:///" + (String)e.nextElement() + url;
                }
            }
            // System.out.println("Open file " + url);
            con = (FileConnection)Connector.open(url);
            // If no exception is thrown, then the URI is valid, but the file may or may not exist.
            if (!con.exists()) { 
                if (create) { 
                    con.create();  // create the file if it doesn't exist
                } else { 
                    return false;
                }
            }

            OutputStream out = con.openOutputStream();
            if (con.fileSize() == 0) { 
                memory = new MemoryHole(SIZE_OF_DB_HDR, Integer.MAX_VALUE);
                byte[] dbHdr = new byte[SIZE_OF_DB_HDR];
                out.write(dbHdr);                
                objToHdr = new ObjToHdr();
            } else { 
                memory = new MemoryHole((int)con.fileSize(), Integer.MAX_VALUE);
                FileObjectInputStream input = new FileObjectInputStream();
                root = input.getRootObject(this);
            }
            out.close();
            output = new FileObjectOutputStream(this);
            opened = true;
        } catch (IOException x) { 
            throw new SermeException("Failed to open storage: " + x);
        }
        return true;
    }

    int getClassId(Class c) { 
        ObjectHeader hdr = objToHdr.get(c);
        int cid;
        if (hdr == null) { 
            cid = ++lastOid;
            String className = c.getName();
            int size = align(SIZE_OF_OBJ_HDR + 2 + className.length()*2);            
            byte[] data = new byte[size];
            try { 
                int pos = allocate(size);
                Converter.packInt(data, OBJ_HDR_SIZE, size);
                Converter.packInt(data, OBJ_HDR_OID, cid);
                Converter.packString(data, SIZE_OF_OBJ_HDR, className);
                OutputStream out = con.openOutputStream(pos);
                out.write(data);
                out.close();
                objToHdr.put(c, cid, pos, size);
            } catch (IOException x) { 
                throw new SermeException("Failed to store class descriptor: " + x);
            }
        } else { 
            cid = hdr.oid;
        }
        return cid;
    }


    public void close() { 
        checkIfOpened(true);
        root = null;
        opened = false;
    }
        
    public Object getRoot() { 
        checkIfOpened(true);
        return root;
    }

    public void setRoot(Object root) { 
        checkIfOpened(true);
        output.storeRoot(root);
        this.root = root;
    }

    public Enumeration enumerateObjects() { 
        return objToHdr.keys();
    }

    void addHole(int offs, int size) {
        if (memory.prev.offs + memory.prev.size == offs) { 
            memory.prev.size += size;
        } else { 
            new MemoryHole(offs, size, memory);
        }
    }

    static class MemoryHole { 
        MemoryHole next;
        MemoryHole prev;
        int        offs;
        int        size;

        void unlink() { 
            next.prev = prev;
            prev.next = next;
        }

        MemoryHole(int offs, int size) { 
            this.offs = offs;
            this.size = size;
            next = prev = this;
        }

        MemoryHole(int offs, int size, MemoryHole head) { 
            this.offs = offs;
            this.size = size;
            prev = head.prev;
            prev.next = this;
            head.prev = this;
            next = head;
        }        
    }


    MemoryHole     memory;
    ObjToHdr       objToHdr;
    Object         root;
    boolean        opened;
    String         path;
    int            lastOid;
    int            rootOid;
    FileConnection con;
    FileObjectOutputStream output;
}