package org.garret.serme.impl;

import java.util.*;
import org.garret.serme.*;

public abstract class ObjectDataInputStream implements IInputStream {
    protected abstract Object getObject(int oid);
    protected abstract Object deserializeObject();

    protected void check(int type) { 
        if (arr[pos++] != type) { 
            throw new SermeException("Expected field type " + Types.MNEM[type] + ", actual type " + Types.MNEM[arr[pos-1]]);
        } 
    }

    public boolean readBoolean() { 
        check(Types.BOOLEAN);
        return arr[pos++] != 0;
    }

    public byte readByte() { 
        check(Types.BYTE);
        return arr[pos++];
    }

    public char readChar() {
        check(Types.CHAR);
        char v = Converter.unpackChar(arr, pos);
        pos += 2;
        return v;
    }

    public short readShort() { 
        check(Types.SHORT);
        short v = Converter.unpackShort(arr, pos);
        pos += 2;
        return v;
    }

    public int readInt() { 
        check(Types.INT);
        int v = Converter.unpackInt(arr, pos);
        pos += 4;
        return v;
    }

    public long readLong() { 
        check(Types.LONG);
        long v = Converter.unpackLong(arr, pos);
        pos += 8;
        return v;
    }

    public String readString() { 
        check(Types.STRING);
        String v = Converter.unpackString(arr, pos);
        pos += 2;
        if (v != null) { 
            pos += v.length()*2;
        }
        return v;
    }

    public Date readDate() { 
        check(Types.DATE);
        Date v = Converter.unpackDate(arr, pos);
        pos += 8;
        return v;
    }

        
    public  boolean[] readArrayOfBoolean() { 
        check(Types.ARRAY_OF_BOOLEAN);
        int size = Converter.unpackInt(arr, pos);
        pos += 4;
        if (size < 0) { 
            return null;
        }
        boolean[] v = new boolean[size];
        int j = pos;
        for (int i = 0; i < size; i++) { 
            v[i] = arr[j++] != 0;
        }
        pos = j;
        return v;
    }
        
    public byte[] readArrayOfByte() {
        check(Types.ARRAY_OF_BYTE);
        int size = Converter.unpackInt(arr, pos);
        pos += 4;
        if (size < 0) { 
            return null;
        }
        byte[] v = new byte[size];
        System.arraycopy(arr, pos, v, 0, size);
        pos += size;
        return v;
    }
        

    public char[] readArrayOfChar() {
        check(Types.ARRAY_OF_CHAR);
        int size = Converter.unpackInt(arr, pos);
        pos += 4;
        if (size < 0) { 
            return null;
        }
        char[] v = new char[size];
        int j = pos;
        for (int i = 0; i < size; i++) { 
            v[i] = Converter.unpackChar(arr, j);
            j += 2;
        }
        pos = j;
        return v;
    }
       
    public short[] readArrayOfShort() { 
        check(Types.ARRAY_OF_SHORT);
        int size = Converter.unpackInt(arr, pos);
        pos += 4;
        if (size < 0) { 
            return null;
        }
        short[] v = new short[size];
        int j = pos;
        for (int i = 0; i < size; i++) { 
            v[i] = Converter.unpackShort(arr, j);
            j += 2;
        }
        pos = j;
        return v;
    }
        
    public int[] readArrayOfInt() {
        check(Types.ARRAY_OF_INT);
        int size = Converter.unpackInt(arr, pos);
        pos += 4;
        if (size < 0) { 
            return null;
        }
        int[] v = new int[size];
        int j = pos;
        for (int i = 0; i < size; i++) { 
            v[i] = Converter.unpackInt(arr, j);
            j += 4;
        }
        pos = j;
        return v;
    }
        
    public long[] readArrayOfLong() { 
        check(Types.ARRAY_OF_LONG);
        int size = Converter.unpackInt(arr, pos);
        pos += 4;
        if (size < 0) { 
            return null;
        }
        long[] v = new long[size];
        int j = pos;
        for (int i = 0; i < size; i++) { 
            v[i] = Converter.unpackLong(arr, j);
            j += 8;
        }
        pos = j;
        return v;
    }
 
    public String[] readArrayOfString() { 
        check(Types.ARRAY_OF_STRING);
        int size = Converter.unpackInt(arr, pos);
        pos += 4;
        if (size < 0) { 
            return null;
        }
        String[] v = new String[size];
        int j = pos;
        for (int i = 0; i < size; i++) { 
            String s = Converter.unpackString(arr, j);            
            j += 2;
            v[i] = s;
            if (s != null) { 
                j += s.length()*2;
            }
        }
        pos = j;
        return v;
    }
 
    public Date[] readArrayOfDate() { 
        check(Types.ARRAY_OF_DATE);
        int size = Converter.unpackInt(arr, pos);
        pos += 4;
        if (size < 0) { 
            return null;
        }
        Date[] v = new Date[size];
        int j = pos;
        for (int i = 0; i < size; i++) { 
            v[i] = Converter.unpackDate(arr, j);
            j += 8;
        }
        pos = j;
        return v;
    }
 
    public Object[] readArrayOfObject() { 
        check(Types.ARRAY_OF_OBJECT);
        int size = Converter.unpackInt(arr, pos);
        pos += 4;
        if (size < 0) { 
            return null;
        }
        Object[] v = new Object[size];
        for (int i = 0; i < size; i++) { 
            v[i] = readObject();
        }
        return v;
    }

    protected void unpackObject(Object obj) {
        if (obj instanceof ISerializable) { 
            ((ISerializable)obj).readObject(this);
        } else if (obj instanceof Vector) { 
            Vector v = (Vector)obj;
            int size = Converter.unpackInt(arr, pos);
            pos += 4;
            while (--size >= 0) { 
                v.addElement(readObject());
            }
        } else if (obj instanceof Hashtable) { 
            Hashtable h = (Hashtable)obj;
            int size = Converter.unpackInt(arr, pos);
            pos += 4;
            while (--size >= 0) { 
                Object key = readObject();
                Object value = readObject();
                h.put(key, value);
            }
        } else { 
            throw new SermeException("Bad object type");
        }
    }

    public Object readObject() { 
        switch (arr[pos++]) { 
        case Types.BOOLEAN:
            return new Boolean(arr[pos++] != 0);
        case Types.BYTE:
            return new Byte(arr[pos++]);
        case Types.CHAR:
        {
            Character v = new Character(Converter.unpackChar(arr, pos));
            pos += 2;
            return v;
        }
        case Types.SHORT:
        {
            Short v = new Short(Converter.unpackShort(arr, pos));
            pos += 2;
            return v;
        }
        case Types.INT:
        {
            Integer v = new Integer(Converter.unpackInt(arr, pos));
            pos += 4;
            return v;
        }            
        case Types.LONG:
        {
            Long v = new Long(Converter.unpackLong(arr, pos));
            pos += 8;
            return v;
        }                    
        case Types.STRING :
        {
            String v = Converter.unpackString(arr, pos);
            pos += (1 + v.length())*2;
            return v;
        }                                
        case Types.DATE:
        {
            Date v = Converter.unpackDate(arr, pos);
            pos += 8;
            return v;
        }                    
        case Types.INSTANCE:
            return deserializeObject();
        case Types.OBJECT:
        {
            int oid = Converter.unpackInt(arr, pos);
            pos += 4;
            return getObject(oid);
        }
        case Types.ARRAY_OF_BOOLEAN:
            pos -= 1;
            return readArrayOfBoolean();
        case Types.ARRAY_OF_BYTE:
            pos -= 1;
            return readArrayOfByte();
        case Types.ARRAY_OF_SHORT:
            pos -= 1;
            return readArrayOfShort();
        case Types.ARRAY_OF_CHAR:
            pos -= 1;
            return readArrayOfChar();
        case Types.ARRAY_OF_INT:
            pos -= 1;
            return readArrayOfInt();
        case Types.ARRAY_OF_LONG:
            pos -= 1;
            return readArrayOfLong();
        case Types.ARRAY_OF_STRING:
            pos -= 1;
            return readArrayOfString();
        case Types.ARRAY_OF_DATE:
            pos -= 1;
            return readArrayOfDate();            
        case Types.ARRAY_OF_OBJECT:
            pos -= 1;
            return readArrayOfObject();            
        default:
            throw new SermeException("Unexpected field type: " + Types.MNEM[arr[pos-1]]);
        }            
    }

    public void close() { 
    }

    
    protected ObjectDataInputStream(byte[] arr) { 
        this.arr = arr;
    }
 
    protected byte[] arr;
    protected int    pos;
}