package org.garret.serme.rms;

import java.io.*;
import java.util.*;
import javax.microedition.rms.*;

import org.garret.serme.*;
import org.garret.serme.impl.*;

/**
 * Implementation of file on top of RMS.
 * This class allows efficiently store and fetched serialized data.
 * Unlike RMS storage it is not possible to update or remove individual objects.
 * It is possible only to load/save all data.
 * The main idea of introducing this class is very slow speed of fetching individual records
 * in many RMS implementations. At the same time size of RMS record is limited and it is not 
 * possible to store all data in one record. So this class can be used as efficient
 * persistent storage for serialized object tree (org.garret.serme.ObjectInputStream,
 * org.garret.serme.ObjectOutputStream)
 */
public class RmsFile { 
    static final int METADATA_RID = 1;
    static final int DEFAULT_RECORD_SIZE = 4096;

    /**
     * Storage constructor
     * @param name name of the storage (passed to <Code>javax.microedition.rms.RecordStore.openStore</code>)
     * @param recordSize maximal RMS record size. RmsFile will split stored data in several 
     * RMS records with specified size.
     */
    public RmsFile(String name, int recordSize) { 
        this.name = name;
        this.recordSize = recordSize;        
    }

    /**
     * Storage constructor
     * @param name name of the storage (passed to <Code>javax.microedition.rms.RecordStore.openStore</code>)
     */
    public RmsFile(String name) { 
        this(name, DEFAULT_RECORD_SIZE);
    }

    /**
     * Get file input stream starting from the specified poistion 
     * @return file input stream
     */
    public InputStream getInputStream(int pos) {
        checkIfOpened(true);
        if (pos > fileSize) { 
            throw new SermeException("Access beyond end of file");
        }
        return new RmsFileInputStream(pos);
    }

    /**
     * Get file input stream starting from the beginning of the file
     * @return file input stream
     */
    public InputStream getInputStream() { 
        return getInputStream(0);
    }

    /** 
     * Get object input stream
     * @return file object input stream
     */
    public IInputStream getObjectInputStream() {
        checkIfOpened(true);
        return new BufferedObjectInputStream(getInputStream(0));
    }

    /**
     * Truncate content of the file.
     * After execution of this method file size is zero
     */
    public void reset() { 
        checkIfOpened(true);
        setSize(0);
    }

    /**
     * Write serialized object tree in the file. This method replace old content of the file 
     * with new one.
     * @param out object output stream
     */
    public void write(IOutputStream out) { 
        checkIfOpened(true);
        fileSize = 0;
        byte[] arr = out.toByteArray();
        OutputStream fs = getOutputStream();
        try { 
            fs.write(arr);
            fs.close();
        } catch (IOException x) { 
            throw new SermeException("Failed to write data: " + x);
        }
    }
    
    /**
     * Get output stream for writting data starting from the specified position
     * @return file output stream
     */
    public OutputStream getOutputStream(int pos) { 
        checkIfOpened(true);
        if (pos > fileSize) { 
            throw new SermeException("Access beyond end of file");
        }
        return new RmsFileOutputStream(pos);
    }

    /**
     * Get output stream for writting data starting from the beginning of the file
     * @return file output stream
     */
    public OutputStream getOutputStream() { 
        return getOutputStream(0);
    }

    /**
     * Get output stream for appending data to the file
     * @return file output stream
     */
    public OutputStream append() { 
        return getOutputStream(fileSize);
    }

    /**
     * Open storage
     * @return true is if storage is sucessfully opened
     */
    public boolean open() { 
        return open0(false);
    }

   /**
     * Create storage if it doesn't exist
     * @return true is if storage is sucessfully created
     */
     public boolean create() { 
        return open0(true);
    }

    /**
     * Delete record storage
     */

    public boolean remove() { 
        checkIfOpened(false);
        try { 
            RecordStore.deleteRecordStore(name);
        } catch (RecordStoreNotFoundException x) { 
            return false;
        } catch (RecordStoreException x) { 
            throw new SermeException("Failed to delete storage: " + x);
        }
        return true;
    }

    /**
     * Close file storage
     */
    public void close() { 
        checkIfOpened(true);
        try { 
            store.closeRecordStore();
        } catch (RecordStoreException x) { 
            throw new SermeException("Failed to close storage: " + x);
        }
        opened = false;
    }
     
    /**
     * Get file size
     * @return  size of the file
     */
    public int size() { 
        return fileSize;
    }


    class RmsFileInputStream extends InputStream { 
        int    recordOffs;
        int    rid;
        int    pos;
        byte[] record;

        RmsFileInputStream(int pos) { 
            rid = METADATA_RID + 1 + (pos / recordSize);
            recordOffs = pos % recordSize;
            this.pos = pos;
            if (pos < fileSize) { 
                try {
                    record = store.getRecord(rid);
                } catch (RecordStoreException x) { 
                    throw new SermeException("Failed to read record: " + x);
                }
            }
        }

        public int read() { 
            byte[] buf = new byte[1];
            return read(buf, 0, 1) == 1 ? (buf[0] & 0xFF) : -1;
        }

        public int available() { 
            return fileSize - pos;
        }

        public int read(byte[] b, int off, int len) { 
            int size = len;
            while (len > 0 && pos < fileSize) {
                if (recordOffs == recordSize) { 
                    try {
                        record = store.getRecord(++rid);
                    } catch (RecordStoreException x) { 
                        throw new SermeException("Failed to read record: " + x);
                    }
                    recordOffs = 0;
                } 
                int n = recordSize - recordOffs < len ? recordSize - recordOffs : len;
                if (n > fileSize - pos) { 
                    n = fileSize - pos;
                }
                System.arraycopy(record, recordOffs, b, off, n);
                off += n;
                len -= n;
                recordOffs += n;
                pos += n;
            }
            return size - len;
        }
    }

    class RmsFileOutputStream extends OutputStream { 
        int    recordOffs;
        int    rid;
        int    pos;
        byte[] record;

        RmsFileOutputStream(int pos) { 
            rid = METADATA_RID + 1 + (pos / recordSize);
            recordOffs = pos % recordSize;
            if (recordOffs != 0) { 
                try {
                    record = store.getRecord(rid);
                } catch (RecordStoreException x) { 
                    throw new SermeException("Failed to read record: " + x);
                }
            } else { 
                record = new byte[recordSize];
            }
        }

        public void write(int b) { 
            byte[] buf = new byte[1];
            buf[0] = (byte)b;
            write(buf, 0, 1);
        }

        public void write(byte[] b, int off, int len) { 
            try {
                if (len != 0) { 
                    do {
                        if (recordOffs == recordSize) { 
                            if (rid > lastRecord) {
                                store.addRecord(record, 0, recordSize);
                                lastRecord = rid;
                            } else { 
                                store.setRecord(rid, record, 0, recordSize);
                            }
                            rid += 1;
                            recordOffs = 0;
                        } 
                        int n = recordSize - recordOffs < len ? recordSize - recordOffs : len;
                        System.arraycopy(b, off, record, recordOffs, n);
                        off += n;
                        len -= n;
                        recordOffs += n;
                        pos += n;
                    } while (len != 0);
                    
                    if (rid > lastRecord) {
                        store.addRecord(record, 0, recordSize);
                        lastRecord = rid;
                    } else { 
                        store.setRecord(rid, record, 0, recordSize);
                    }                
                    if (pos > fileSize) { 
                        setSize(pos);
                    } 
                }
            } catch (RecordStoreException x) { 
                throw new SermeException("Failed to write record: " + x);
            }
        }
    }

    protected void setSize(int size) { 
        fileSize = size;
        byte[] metadata = new byte[4];
        Converter.packInt(metadata, 0, size);
        try { 
            store.setRecord(METADATA_RID, metadata, 0, 4);
        } catch (RecordStoreException x) { 
            throw new SermeException("Failed tp update metadata: " + x);
        }
    }

    protected void checkIfOpened(boolean shouldBeOpened) { 
        if (opened != shouldBeOpened) { 
            throw new SermeException("Incorrect storage state");
        }
    }

    protected boolean open0(boolean create) {         
        checkIfOpened(false);
        try { 
            store = RecordStore.openRecordStore(name, create);
            int size = store.getNumRecords();
            if (size == 0) { 
                byte[] metadata = new byte[4];
                store.addRecord(metadata, 0, 4);
                fileSize = 0;
                lastRecord = METADATA_RID;
            } else { 
                byte[] metadata = store.getRecord(METADATA_RID);
                fileSize = Converter.unpackInt(metadata, 0);
                lastRecord = size;
            }
            opened = true;
        } catch (RecordStoreNotFoundException x) { 
            return false;
        } catch (RecordStoreException x) { 
            throw new SermeException("Failed to open storage: " + x);
        }
        return true;
    }

    RecordStore  store;
    boolean      opened;
    String       name;
    int          fileSize;
    int          recordSize;
    int          lastRecord;
}